#include "MainWindow.h"
#include "Aux_Transform_Mesh_Dialog.h"

VolumeMeshProcessing::VolumeMeshProcessing()
{
	initMainWindow();
}

VolumeMeshProcessing::~VolumeMeshProcessing()
{

}

void VolumeMeshProcessing::initMainWindow()
{
	createCenterWidget();
	createAction();
	CreateMenu();
	createContextMenu();
	createToolBar();
	createStatusBar();

	latestDrawMode = MeshViewerWidget::FLAT_LINES;

	connect(center,SIGNAL(setMouseMode_signal_main(int)),this,SLOT(setMouseMode_slot(int)));
	connect(center,SIGNAL(setDrawMode_signal_main(int)),this,SLOT(setDrawMode_slot(int)));
}

void VolumeMeshProcessing::createCenterWidget()
{
	center = new centerWidget();
	this->setCentralWidget(center);

	//center->setAcceptDrops(true);
	this->setAcceptDrops(false);
}

void VolumeMeshProcessing::createAction()
{
	//file action
	openAction = new QAction(("&Open"), this);
	openAction->setIcon(QIcon(":/VolumeMeshProcessing/Images/Open.png"));
	openAction->setShortcut(QKeySequence::Open);
	openAction->setStatusTip(("Open a mesh file"));
	connect(openAction, SIGNAL(triggered()), center, SLOT(open_mesh_query()));

	saveAction = new QAction(("&Save"), this);
	saveAction->setIcon(QIcon(":/VolumeMeshProcessing/Images/Save.png"));
	saveAction->setShortcut(QKeySequence::Save);
	saveAction->setStatusTip(("Save the mesh to file"));
	connect(saveAction, SIGNAL(triggered()), center, SLOT(save_mesh_query()));

	exitAction = new QAction(("E&xit"), this);
	exitAction->setShortcut(Qt::CTRL + Qt::Key_Q);
	exitAction->setStatusTip(("Exit the application"));
	connect(exitAction, SIGNAL(triggered()), this, SLOT(close()));

	clearAction = new QAction("Clear ALL",this);
	clearAction->setIcon(QIcon(":/VolumeMeshProcessing/Images/ClearMesh.png"));
	clearAction->setShortcut(Qt::CTRL + Qt::Key_W);
	clearAction->setStatusTip(("Clear all Mesh"));
	connect(clearAction, SIGNAL(triggered()), center, SLOT(clear_all_mesh()));

	//mesh viewer
	wireFrameAction = new QAction(("&WireFrame"), this);
	wireFrameAction->setIcon(QIcon(":/VolumeMeshProcessing/Images/wire.png"));
	wireFrameAction->setStatusTip(("Using wireFrame showing method"));
	wireFrameAction->setCheckable(true);
	wireFrameAction->setChecked(false);
	connect(wireFrameAction, SIGNAL(triggered()), this, SLOT(wireFrameShow()));

	solidflatAction = new QAction("Solid Flat", this);
	solidflatAction->setIcon(QIcon(":/VolumeMeshProcessing/Images/flat.png"));
	solidflatAction->setStatusTip(("Using Solid Falt showing method"));
	solidflatAction->setCheckable(true);
	solidflatAction->setChecked(false);
	connect(solidflatAction, SIGNAL(triggered()), this, SLOT(solidflatShow()));

	flatLinesAction = new QAction("Flat&Lines", this);
	flatLinesAction->setIcon(QIcon(":/VolumeMeshProcessing/Images/flatlines.png"));
	flatLinesAction->setStatusTip(("Using flatLines showing method"));
	flatLinesAction->setCheckable(true);
	flatLinesAction->setChecked(true);
	connect(flatLinesAction, SIGNAL(triggered()), this, SLOT(flatLinesShow()));

	allLinesAction = new QAction("All&Lines", this);
	allLinesAction->setIcon(QIcon(":/VolumeMeshProcessing/Images/all_lines.png"));
	allLinesAction->setStatusTip(("Show All Lines"));
	allLinesAction->setCheckable(true);
	allLinesAction->setChecked(false);
	connect(allLinesAction, SIGNAL(triggered()), this, SLOT(allLinesShow()));
	
	allPointsAction = new QAction("ALL Points", this);
	allPointsAction->setIcon(QIcon(":/VolumeMeshProcessing/Images/points.png"));
	allPointsAction->setStatusTip(("Show All Points"));
	allPointsAction->setCheckable(true);
	allPointsAction->setChecked(false);
	connect(allPointsAction, SIGNAL(triggered()), this, SLOT(allPointsShow()));

	allPolyAction = new QAction("All&Poly", this);
	allPolyAction->setIcon(QIcon(":/VolumeMeshProcessing/Images/all_poly.png"));
	allPolyAction->setStatusTip(("Using AllLines showing method"));
	allPolyAction->setCheckable(true);
	allPolyAction->setChecked(false);
	connect(allPolyAction, SIGNAL(triggered()), this, SLOT(allPolysShow()));

	SingularityAction = new QAction("Singularity",this);
	SingularityAction->setIcon(QIcon(":/VolumeMeshProcessing/Images/singularity.png"));
	SingularityAction->setStatusTip(("Using AllLines showing method"));
	SingularityAction->setCheckable(true);
	SingularityAction->setChecked(false);
	connect(SingularityAction, SIGNAL(triggered()), this, SLOT(singularityShow()));

	textureAction = new QAction("Texture", this);
	textureAction->setIcon(QIcon(":/VolumeMeshProcessing/Images/texture_3D.png"));
	textureAction->setStatusTip(("Texture Show"));
	textureAction->setCheckable(true);
	textureAction->setChecked(false);
	connect(textureAction, SIGNAL(triggered()), this, SLOT(textureShow()));

	//pick
	pickCellAction = new QAction(("Pick Cell"), this);
	pickCellAction->setIcon(QIcon(":/VolumeMeshProcessing/Images/select_cell.png"));
	pickCellAction->setStatusTip(("Pick Cell"));
	pickCellAction->setCheckable(true);
	pickCellAction->setChecked(false);
	pickCellAction->setEnabled(false);
	connect(pickCellAction, SIGNAL(triggered()), this, SLOT(pickCellMode()));

	pickVolumeAction = new QAction(("Pick Volume"), this);
	pickVolumeAction->setIcon(QIcon(":/VolumeMeshProcessing/Images/select_volume.png"));
	pickVolumeAction->setStatusTip(("Pick Volume"));
	pickVolumeAction->setCheckable(true);
	pickVolumeAction->setChecked(false);
	connect(pickVolumeAction, SIGNAL(triggered()), this, SLOT(pickVolumeMode()));

	pickSheetAction = new QAction(("Pick Sheet"), this);
	pickSheetAction->setIcon(QIcon(":/VolumeMeshProcessing/Images/select_sheet.png"));
	pickSheetAction->setStatusTip(("Pick Sheet"));
	pickSheetAction->setCheckable(true);
	pickSheetAction->setChecked(false);
	connect(pickSheetAction, SIGNAL(triggered()), this, SLOT(pickSheetMode()));

	pickVertexAction = new QAction(("Pick Vertex"), this);
	pickVertexAction->setIcon(QIcon(":/VolumeMeshProcessing/Images/select_vertex.png"));
	pickVertexAction->setStatusTip(("Pick Vertex"));
	pickVertexAction->setCheckable(true);
	pickVertexAction->setChecked(false);
	connect(pickVertexAction, SIGNAL(triggered()), this, SLOT(pickVertexMode()));

	moveVertexAction = new QAction(("Move Vertex"), this);
	moveVertexAction->setIcon(QIcon(":/VolumeMeshProcessing/Images/move_vertex.png"));
	moveVertexAction->setStatusTip(("Move Vertex"));
	moveVertexAction->setCheckable(true);
	moveVertexAction->setChecked(false);
	connect(moveVertexAction, SIGNAL(triggered()), this, SLOT(moveVertexMode()));

	pickPointAction = new QAction(("Pick Point"), this);
	pickPointAction->setIcon(QIcon(":/VolumeMeshProcessing/Images/select_point.png"));
	pickPointAction->setStatusTip(("Pick Point"));
	pickPointAction->setCheckable(true);
	pickPointAction->setChecked(false);
	connect(pickPointAction, SIGNAL(triggered()), this, SLOT(pickPointMode())); 

	pickFaceAction = new QAction(("Pick Face"), this);
	pickFaceAction->setIcon(QIcon(":/VolumeMeshProcessing/Images/select_face.png"));
	pickFaceAction->setStatusTip(("Pick Face"));
	pickFaceAction->setCheckable(true);
	pickFaceAction->setChecked(false);
	connect(pickFaceAction, SIGNAL(triggered()), this, SLOT(pickFaceMode())); 

	pickEdgeAction = new QAction(("Pick Edge"), this);
	pickEdgeAction->setIcon(QIcon(":/VolumeMeshProcessing/Images/select_edge.png"));
	pickEdgeAction->setStatusTip(("Pick Edge"));
	pickEdgeAction->setCheckable(true);
	pickEdgeAction->setChecked(false);
	connect(pickEdgeAction, SIGNAL(triggered()), this, SLOT(pickEdgeMode()));

	saveScreenAction = new QAction("Save Screen",this);
	saveScreenAction->setIcon(QIcon(":/VolumeMeshProcessing/Images/saveScreen.png"));
	saveScreenAction->setStatusTip(("Save Screen"));
	saveScreenAction->setCheckable(true);
	saveScreenAction->setChecked(false);
	connect(saveScreenAction, SIGNAL(triggered()), this, SLOT(saveScreen()));

	subdivideTetraAction = new QAction("Subdivide Tet",this);
	subdivideTetraAction->setStatusTip(("Subdivide Tetrahedral"));
	connect(subdivideTetraAction, SIGNAL(triggered()), this, SLOT(subdivideTetra()));

	transform_mesh_Action = new QAction("Transform Mesh", this);
	transform_mesh_Action->setStatusTip(("Transform Mesh"));
	connect(transform_mesh_Action, SIGNAL(triggered()), this, SLOT(transform_mesh()));

	create_grid_all_hex_mesh_Action = new QAction("Create Grid Hex", this);
	create_grid_all_hex_mesh_Action->setStatusTip(("Create Grid Hex"));
	connect(create_grid_all_hex_mesh_Action, SIGNAL(triggered()), this, SLOT(create_grid_all_hex_mesh()));

	create_grid_mesh_Action = new QAction("Create Grid Mesh", this);
	create_grid_mesh_Action->setStatusTip(("Create Grid Mesh"));
	connect(create_grid_mesh_Action, SIGNAL(triggered()), this, SLOT(create_grid_mesh()));
}

void VolumeMeshProcessing::CreateMenu()
{
	fileMenu = menuBar()->addMenu("File");
	fileMenu->addAction(openAction);
	fileMenu->addAction(saveAction);
	fileMenu->addAction(clearAction);
	fileMenu->addSeparator();
	fileMenu->addAction(saveScreenAction);
	fileMenu->addSeparator();
	fileMenu->addAction(exitAction);

	viewMenu = menuBar()->addMenu("View");
	viewMenu->addAction(wireFrameAction);
	viewMenu->addAction(solidflatAction);
	viewMenu->addAction(flatLinesAction);
	viewMenu->addAction(allLinesAction);
	viewMenu->addAction(allPointsAction);
	viewMenu->addAction(allPolyAction);
	viewMenu->addAction(SingularityAction);
	viewMenu->addAction(textureAction);
	viewMenu->addSeparator();

	pickMenu = menuBar()->addMenu("Pick");
	pickMenu->addAction(pickCellAction);
	pickMenu->addAction(pickVolumeAction);
	pickMenu->addAction(pickSheetAction);
	pickMenu->addAction(pickVertexAction);
	pickMenu->addAction(moveVertexAction);
	pickMenu->addAction(pickPointAction);
	pickMenu->addAction(pickFaceAction);
	pickMenu->addAction(pickEdgeAction);
	pickMenu->addSeparator();

	AuxMenu = menuBar()->addMenu("Auxiliary");
	AuxMenu->addAction(subdivideTetraAction);
	AuxMenu->addAction(transform_mesh_Action);
	AuxMenu->addAction(create_grid_all_hex_mesh_Action);
	AuxMenu->addAction(create_grid_mesh_Action);
}

void VolumeMeshProcessing::createContextMenu()
{

}

void VolumeMeshProcessing::createToolBar()
{
	fileToolBar = addToolBar(("&File"));
	fileToolBar->addAction(openAction);
	fileToolBar->addAction(saveAction);
	fileToolBar->addAction(clearAction);
	fileToolBar->addAction(saveScreenAction);

	viewToolBar = addToolBar(("&View"));
	viewToolBar->addAction(wireFrameAction);
	viewToolBar->addAction(solidflatAction);
	viewToolBar->addAction(flatLinesAction);
	viewToolBar->addAction(allLinesAction);
	viewToolBar->addAction(allPointsAction);
	viewToolBar->addAction(allPolyAction);
	viewToolBar->addAction(SingularityAction);
	viewToolBar->addAction(textureAction);

	pickToolBar = addToolBar("Pick");
	pickToolBar->addAction(pickCellAction);
	pickToolBar->addAction(pickVolumeAction);
	pickToolBar->addAction(pickSheetAction);
	pickToolBar->addAction(pickVertexAction);
	pickToolBar->addAction(moveVertexAction);
	pickToolBar->addAction(pickPointAction);
	pickToolBar->addAction(pickFaceAction);
	pickToolBar->addAction(pickEdgeAction);
}

void VolumeMeshProcessing::createStatusBar()
{
	statusLabel = new QLabel(("no mesh"));
	statusLabel->setAlignment(Qt::AlignHCenter);

	connect(center, SIGNAL( haveLoadMesh(QString) ), statusLabel, SLOT( setText(QString) ) );

	statusBar()->addWidget(statusLabel);
}

void VolumeMeshProcessing::wireFrameShow()
{
	setAllActionChecked(false);
	wireFrameAction->setChecked(true);
	center->setDrawMode(MeshViewerWidget::WIRE_FRAME);
	center->setMouseMode(InteractiveWidget::TRANS);
}

void VolumeMeshProcessing::solidflatShow()
{
	setAllActionChecked(false);
	solidflatAction->setChecked(true);
	center->setDrawMode(MeshViewerWidget::SOLID_FLAT);
	center->setMouseMode(InteractiveWidget::TRANS);
}

void VolumeMeshProcessing::flatLinesShow()
{
	setAllActionChecked(false);
	flatLinesAction->setChecked(true);
	center->setDrawMode(MeshViewerWidget::FLAT_LINES);
	center->setMouseMode(InteractiveWidget::TRANS);
}

void VolumeMeshProcessing::allLinesShow()
{
	setAllActionChecked(false);
	allLinesAction->setChecked(true);
	center->setDrawMode(MeshViewerWidget::ALL_LINES);
	center->setMouseMode(InteractiveWidget::TRANS);
}

void VolumeMeshProcessing::allPointsShow()
{
	setAllActionChecked(false);
	allPointsAction->setChecked(true);
	center->setDrawMode(MeshViewerWidget::ALL_POINTS);
	center->setMouseMode(InteractiveWidget::TRANS);
}

void VolumeMeshProcessing::allPolysShow()
{
	setAllActionChecked(false);
	allPolyAction->setChecked(true);
	center->setDrawMode(MeshViewerWidget::ALL_POLYHEDRALS);
	center->setMouseMode(InteractiveWidget::TRANS);
}

void VolumeMeshProcessing::singularityShow()
{
	setAllActionChecked(false);
	SingularityAction->setChecked(true);
	center->setDrawMode(MeshViewerWidget::SINGULARITY);
	center->setMouseMode(InteractiveWidget::TRANS);
}

void VolumeMeshProcessing::textureShow()
{
	setAllActionChecked(false);
	textureAction->setChecked(true);
	center->setDrawMode(MeshViewerWidget::VOLUME_TEXTURE);
	center->setMouseMode(InteractiveWidget::TRANS);
}

void VolumeMeshProcessing::pickCellMode()
{
	setAllActionChecked(false);
	pickCellAction->setChecked(true);
	center->setMouseMode(InteractiveWidget::CELLPICK);
}

void VolumeMeshProcessing::pickVolumeMode()
{
	setAllActionChecked(false);
	pickVolumeAction->setChecked(true);
	center->setMouseMode(InteractiveWidget::VolumePICK);
}

void VolumeMeshProcessing::pickSheetMode()
{
	setAllActionChecked(false);
	pickSheetAction->setChecked(true);
	center->setMouseMode(InteractiveWidget::SHEETPICK);
}

void VolumeMeshProcessing::pickVertexMode()
{
	setAllActionChecked(false);
	pickVertexAction->setChecked(true);
	center->setMouseMode(InteractiveWidget::VERTEXPICK);
}

void VolumeMeshProcessing::moveVertexMode()
{
	setAllActionChecked(false);
	moveVertexAction->setChecked(true);
	center->setMouseMode(InteractiveWidget::VERTEXMOVE);
}

void VolumeMeshProcessing::pickPointMode()
{
	setAllActionChecked(false);
	pickPointAction->setChecked(true);
	center->setMouseMode(InteractiveWidget::POINTPICK);
	//center->set_inserting_lines(false);
}

void VolumeMeshProcessing::pickFaceMode()
{
	setAllActionChecked(false);
	pickFaceAction->setChecked(true);
	center->setMouseMode(InteractiveWidget::FACEPICK);
	//center->set_draw_only_lines();
}

void VolumeMeshProcessing::pickEdgeMode()
{
	setAllActionChecked(false);
	pickEdgeAction->setChecked(true);
	center->setMouseMode(InteractiveWidget::EDGEPICK);
	//center->set_draw_only_lines();
}

void VolumeMeshProcessing::setAllActionChecked(bool b)
{
	//viewer
	wireFrameAction->setChecked(b);
	solidflatAction->setChecked(b);
	flatLinesAction->setChecked(b);
	allLinesAction->setChecked(b);
	allPolyAction->setChecked(b);
	SingularityAction->setChecked(b);
	textureAction->setChecked(b);

	//pick
	pickCellAction->setChecked(b);
	pickVolumeAction->setChecked(b);
	pickSheetAction->setChecked(b);
	pickVertexAction->setChecked(b);
	pickPointAction->setChecked(b);
	pickFaceAction->setChecked(b);
	pickEdgeAction->setChecked(b);
	moveVertexAction->setChecked(b);
}

void VolumeMeshProcessing::setAllViewActionChecked(bool b)
{
	//viewer
	wireFrameAction->setChecked(b);
	solidflatAction->setChecked(b);
	flatLinesAction->setChecked(b);
	allLinesAction->setChecked(b);
	allPointsAction->setChecked(b);
	allPolyAction->setChecked(b);
	SingularityAction->setChecked(b);
	textureAction->setChecked(b);
}

void VolumeMeshProcessing::setAllMouseActionChecked(bool b)
{
	//pick
	pickCellAction->setChecked(b);
	pickVolumeAction->setChecked(b);
	pickSheetAction->setChecked(b);
	pickVertexAction->setChecked(b);
	pickPointAction->setChecked(b);
	pickFaceAction->setChecked(b);
	pickEdgeAction->setChecked(b);
	moveVertexAction->setChecked(b);
}

void VolumeMeshProcessing::saveScreen()
{
	center->saveOpenGLScreen();
}

void VolumeMeshProcessing::subdivideTetra()
{
	center->subdivideTetradetralMesh();
}

void VolumeMeshProcessing::transform_mesh()
{
	aux_transform_mesh_dialog* d = new aux_transform_mesh_dialog();
	int r = d->exec();
	if(r == QDialog::Accepted)
	{
		std::vector<double> tm; d->get_transform_matrix(tm);
		center->transform_mesh(tm);
	}
	else if(r == QDialog::Rejected)
	{

	}
	delete d;
}

void VolumeMeshProcessing::create_grid_all_hex_mesh()
{
	create_all_hex_grid_mesh_dialog* d = new create_all_hex_grid_mesh_dialog();
	int r = d->exec();
	if (r == QDialog::Accepted)
	{
		int w,h,l; d->get_w_h_l(w,h,l);
		center->create_grid_all_hex_mesh(w, h, l);
	}
	else if (r == QDialog::Rejected)
	{

	}
	delete d;
}

void VolumeMeshProcessing::create_grid_mesh()
{
	center->create_grid_mesh();
}

void VolumeMeshProcessing::setMouseMode_slot(int mm)
{
	if(mm == InteractiveWidget::TRANS)
	{
		setAllActionChecked(false);
		setAllViewActionChecked(false);
		switch(latestDrawMode)
		{
		case InteractiveWidget::WIRE_FRAME:
			wireFrameAction->setChecked(true);
			break;
		case InteractiveWidget::SOLID_FLAT:
			solidflatAction->setChecked(true);
			break;
		case InteractiveWidget::FLAT_LINES:
			flatLinesAction->setChecked(true);
			break;
		case InteractiveWidget::ALL_LINES:
			allLinesAction->setChecked(true);
			break;
		case InteractiveWidget::ALL_POLYHEDRALS:
			allPolyAction->setChecked(true);
			break;
		case InteractiveWidget::SINGULARITY:
			SingularityAction->setChecked(true);
			break;
		case InteractiveWidget::VOLUME_TEXTURE:
			textureAction->setChecked(true);
			break;
		}
	}
	else
	{
		setAllViewActionChecked(false);
	}
}

void VolumeMeshProcessing::setDrawMode_slot(int dm)
{
	setAllMouseActionChecked(false);
	latestDrawMode = dm;
}